from pycparser import c_ast as ast

from .Node import Node
from .StructNodeMember import StructNodeMember


class StructNode(Node):
    def __init__(self, decl_node, nested=False):
        self.name = decl_node.declname if isinstance(decl_node, ast.TypeDecl) else decl_node.type.name
        self.members = []
        self.nested = nested

        for decl in decl_node.type.decls:
            if StructNodeMember.is_type(decl.type):
                self.members.append(StructNodeMember(decl.type))

    @staticmethod
    def is_type(decl_node):
        return isinstance(decl_node.type, ast.Struct) and decl_node.type.decls



from pycparser import c_ast as ast

from .Node import Node
from .FunctionNode import FunctionNode
from .Collection import Collection


class FunctionNodeCollection(Node, Collection):
    def __init__(self, top_node):
        self.functions: list[FunctionNode] = []
        for nested in top_node:
            if FunctionNode.is_type(nested):
                self.functions.append(FunctionNode(nested))
            elif FunctionNode.is_type(nested.type):
                self.functions.append(FunctionNode(nested.type))

    @staticmethod
    def is_type(decl_node):
        return isinstance(decl_node, ast.FileAST)



import logging
import os
import re

import pycparser_fake_libc


def extract_model_data(rtmodel_path: str) -> tuple[str, str]:
    with open(rtmodel_path) as file_obj:
        rtmodel_data = file_obj.read()
    return extract_model_name(rtmodel_data), extract_model_version(rtmodel_data)


def extract_model_version(rtmodel_data: str) -> str:
    match = re.search(r'Model version *: ([0-9.]+).*', rtmodel_data)
    if not match:
        logging.warning("Can't detect model version.")
        return "1"
    return match.group(1)


def extract_model_name(rtmodel_data: str) -> str:
    match = re.search(r'#include "(.*?).h".*', rtmodel_data)
    if not match:
        raise ValueError("Can't detect model name")
    return match.group(1)


def collect_includes(dir_path: str) -> list[str]:
    include_paths = [dir_path, pycparser_fake_libc.directory]
    includes = ["-I" + path for path in include_paths]
    return includes


def collect_files_with_ext(dir_path: str, ext: str) -> list[str]:
    result = []
    for file in os.listdir(dir_path):
        if file.endswith(ext):
            result.append(os.path.join(dir_path, file))
    return result



from __future__ import annotations
import asyncio
import os

from pycparser import parse_file, c_ast as ast

from .nodes import StructNodeCollection, FunctionNodeCollection, VariableNodeCollection, TypeNodeMap
from .architect_utils import extract_model_data, collect_includes


class Collector:
    def __init__(self, c_code_path: str, python_package_name: str, time_step: float):
        self.c_code_tmp_path = c_code_path
        self.python_package_name = python_package_name
        self.time_step = time_step

        self.rtmodel_h_path = os.path.join(self.c_code_tmp_path, 'rtmodel.h')
        if not os.path.exists(self.rtmodel_h_path):
            raise ValueError("Can not find 'rtmodel.h'")

        main_c_path = os.path.join(self.c_code_tmp_path, 'rt_main.c')
        if os.path.exists(main_c_path):
            os.remove(main_c_path)

        self.model_name, self.model_version = extract_model_data(self.rtmodel_h_path)

        includes = collect_includes(self.c_code_tmp_path)
        top_node = parse_file(self.rtmodel_h_path, use_cpp=True, cpp_args=includes)
        if not isinstance(top_node, ast.FileAST):
            raise ValueError("Didn't find valid declaration after parsing header.")

        self.structs_collection = StructNodeCollection(top_node)
        self.functions_collection = FunctionNodeCollection(top_node)
        self.variables_collection = VariableNodeCollection(top_node)
        self.types_map = TypeNodeMap(top_node)

    @property
    def input_members(self) -> dict:
        return self.get_members(f"ExtU_{self.model_name}_T")

    @property
    def output_members(self) -> dict:
        return self.get_members(f"ExtY_{self.model_name}_T")

    def get_members(self, struct_name: str) -> dict:
        return {
            member.name: self.types_map[member.type]
            for member in self.structs_collection[struct_name].members
        } if struct_name in self.structs_collection else {}

    @classmethod
    async def create(cls, c_code_path: str, python_package_name: str, time_step: float) -> Collector:
        return await asyncio.to_thread(cls, c_code_path, python_package_name, time_step)



import os
import re
import sys
import shutil
import logging
import asyncio
import argparse
import tempfile

from simbind.architect.simbind_architect import Collector, MockDriver, SiLDriver, test_model_integrity
from simbind.exporter.simulink_exporter import export_model, FixedStepSolver


def dir_has_files(path: str) -> bool:
    if not os.path.isdir(path):
        raise RuntimeError(f"'{path}' is not a directory")
    return len(os.listdir(path)) != 0


def delete_dir_internals(path: str):
    for filename in os.listdir(path):
        file_path = os.path.join(path, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)


def prepare_dir(path: str, overwrite: bool):
    if dir_has_files(path):
        if not overwrite:
            raise ValueError(f"Directory {path} not empty. Consider using --overwrite.")
        delete_dir_internals(path)


async def assemble(
        slx_path: str,
        package_name: str,
        exporter_out_dir: str,
        models_out_dir: str,
        wheel_out_dir: str,
        solver: str,
        time_step: float,
        license_text: str
):
    await asyncio.to_thread(
        export_model,
        slx_path,
        exporter_out_dir,
        time_step,
        FixedStepSolver(solver)
    )
    collector = await Collector.create(exporter_out_dir, package_name, time_step)

    mock_dir = os.path.join(models_out_dir, "mock")
    os.mkdir(mock_dir)
    mock_pkg_name, mock_cls_name = await MockDriver.compose(collector, mock_dir, license_text)

    sil_dir = os.path.join(models_out_dir, "sil")
    os.mkdir(sil_dir)
    sil_pkg_name, sil_cls_name = await SiLDriver.compose(collector, sil_dir, license_text)

    await test_model_integrity(
        mock_dir,
        mock_pkg_name,
        mock_cls_name,
        collector,
    )

    await test_model_integrity(
        sil_dir,
        sil_pkg_name,
        sil_cls_name,
        collector
    )

    await test_model_integrity(
        sil_dir,
        sil_pkg_name,
        sil_cls_name,
        collector,
        True
    )

    pip_proc = await asyncio.create_subprocess_shell(
        f"{sys.executable} -m pip wheel --no-deps {sil_dir} -w {wheel_out_dir}",
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.PIPE
    )
    _, stderr = await pip_proc.communicate()
    if pip_proc.returncode:
        logging.error(stderr)
        raise RuntimeError(stderr)


async def async_main():
    parser = argparse.ArgumentParser(
        prog='simbind',
        description='Tool to generate Python wheel package from Simulink model.',
    )
    parser.add_argument(
        '--slx-path',
        dest='slx_path',
        help='Path to Simulink .slx file.',
        required=True,
        type=os.path.abspath
    )
    parser.add_argument(
        '--pkg-name',
        dest='pkg_name',
        help='Name of the output Python package. (default: \'%(default)s\')',
        type=str,
        default='model'
    )
    parser.add_argument(
        '--wheel-out-dir',
        dest='wheel_out_dir',
        help='Path to folder where wheel package will be stored. (default: \'%(default)s\')',
        default='.',
        type=os.path.abspath
    )
    parser.add_argument(
        '--solver',
        dest='solver',
        choices=['ode1', 'ode2', 'ode3', 'ode4', 'ode5', 'ode8', 'ode14x', 'ode1be'],
        help='Fixed-step solver. (default: \'%(default)s\')',
        type=str,
        default='ode5'
    )
    parser.add_argument(
        '--step',
        dest='step_size',
        help='Fixed step size in seconds. (default: \'%(default)s\')',
        type=float,
        default=0.001
    )
    parser.add_argument(
        '--license-text',
        dest='license_text',
        help='License text that will be included in output Python wheel package. (default: \'%(default)s\')',
        type=str,
        default=""
    )
    parser.add_argument(
        '-v',
        dest='verbosity',
        default=0,
        help='Specifies the level of verbosity. Example: -vvv',
        action='count'
    )
    # [DEBUG] Path to directory that will be used to store .c and .h files exported by Simulink
    parser.add_argument(
        '--exporter-out-dir',
        dest='exporter_out_dir',
        help=argparse.SUPPRESS,
        type=os.path.abspath
    )
    # [DEBUG] Directory where output Python model will be stored
    parser.add_argument(
        '--models-out-dir',
        dest='models_out_dir',
        help=argparse.SUPPRESS,
        type=os.path.abspath
    )
    # [DEBUG] Flag to delete models-out-dir and exporter-out-dir folders' content if they are not empty
    parser.add_argument(
        '--overwrite',
        dest='overwrite',
        help=argparse.SUPPRESS,
        action='store_true'
    )

    args = parser.parse_args()

    verbosity2lvl = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG
    }
    args.verbosity = 3 if args.verbosity > 3 else args.verbosity
    logger = logging.getLogger()
    logger.setLevel(level=verbosity2lvl[args.verbosity])
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt='%(levelname)-8s :: %(message)s')
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    exporter_out_tmp_dir: tempfile.TemporaryDirectory | None = None
    models_out_tmp_dir: tempfile.TemporaryDirectory | None = None

    if not os.path.isfile(args.slx_path):
        raise FileNotFoundError(f"File '{args.slx_path}' does not exist.")

    if not os.path.isdir(args.wheel_out_dir):
        raise FileNotFoundError(f"Directory '{args.wheel_out_dir}' does not exist.")

    if not re.match(r"^[a-z][a-z0-9_]{4,29}$", args.pkg_name):
        raise ValueError(
            f"Invalid package name '{args.pkg_name}'."
            "Only lowercase letters, numbers and underscores are acceptable. Should start with lowercase letter and "
            "have length between 5 and 50 characters."
        )

    try:
        if args.exporter_out_dir:
            prepare_dir(args.exporter_out_dir, args.overwrite)
        else:
            exporter_out_tmp_dir = tempfile.TemporaryDirectory()

        if args.models_out_dir:
            prepare_dir(args.models_out_dir, args.overwrite)
        else:
            models_out_tmp_dir = tempfile.TemporaryDirectory()

        await assemble(
            args.slx_path,
            args.pkg_name,
            args.exporter_out_dir or exporter_out_tmp_dir.name,
            args.models_out_dir or models_out_tmp_dir.name,
            args.wheel_out_dir,
            args.solver,
            args.step_size,
            args.license_text
        )
    finally:
        if exporter_out_tmp_dir:
            exporter_out_tmp_dir.cleanup()
        if models_out_tmp_dir:
            models_out_tmp_dir.cleanup()


def main():
    asyncio.run(async_main())


if __name__ == '__main__':
    main()



import asyncio
import logging
import os
import tempfile

from jinja2 import Environment, FileSystemLoader

from ..Collector import Collector
from ..Driver import Driver


class SiLDriver(Driver):
    _templ_dir = os.path.abspath(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')
    )
    _env = Environment(loader=FileSystemLoader(_templ_dir), autoescape=True, keep_trailing_newline=True)

    @classmethod
    async def compose(cls, collector: Collector, output_dir: str, license_text: str) -> tuple[str, str]:
        if not os.path.isdir(output_dir):
            raise ValueError(f"Directory '{output_dir}' not found.")

        pkg_dir = os.path.join(output_dir, collector.python_package_name)
        os.mkdir(pkg_dir)

        build_dir = tempfile.TemporaryDirectory()
        stash_dir = tempfile.TemporaryDirectory()
        try:
            init_str = cls._env.get_template("__init__.py.j2").render(
                model_name=collector.model_name
            )
            init_path = os.path.join(pkg_dir, "__init__.py")
            with open(init_path, 'w') as file_obj:
                file_obj.write(init_str)

            base_model_str = cls._env.get_template("BaseModel.py.j2").render()
            base_model_path = os.path.join(pkg_dir, "BaseModel.py")
            with open(base_model_path, 'w') as file_obj:
                file_obj.write(base_model_str)

            model_str = cls._env.get_template("Model.py.j2").render(
                model_name=collector.model_name,
                input_map=collector.input_members,
                output_map=collector.output_members
            )
            model_path = os.path.join(pkg_dir, "Model.py")
            with open(model_path, 'w') as file_obj:
                file_obj.write(model_str)

            pyproject_str = cls._env.get_template("pyproject.toml.j2").render(
                model_version=collector.model_version,
                package_name=collector.python_package_name,
                model_name=collector.model_name
            )
            pyproject_path = os.path.join(output_dir, "pyproject.toml")
            with open(pyproject_path, 'w') as file_obj:
                file_obj.write(pyproject_str)

            license_path = os.path.join(output_dir, "LICENSE")
            with open(license_path, 'w') as file_obj:
                file_obj.write(license_text)

            readme_str = cls._env.get_template("README.md.j2").render()
            readme_path = os.path.join(output_dir, "README.md")
            with open(readme_path, 'w') as file_obj:
                file_obj.write(readme_str)

            swig_str = cls._env.get_template("swig.i.j2").render(
                model_name=collector.model_name
            )
            swig_path = os.path.join(stash_dir.name, "swig.i")
            with open(swig_path, 'w') as file_obj:
                file_obj.write(swig_str)

            cmake_str = cls._env.get_template("CMakeLists.txt.j2").render(
                model_name=collector.model_name,
                source_dir=collector.c_code_tmp_path,
                swig_interface_path=swig_path,
                output_dir=pkg_dir,
                build_dir=build_dir.name,
            )
            cmake_path = os.path.join(stash_dir.name, "CMakeLists.txt")
            with open(cmake_path, 'w') as file_obj:
                file_obj.write(cmake_str)

            cmake_proc = await asyncio.create_subprocess_exec(
                'cmake',
                f'-B{build_dir.name}',
                f'-S{stash_dir.name}',
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.PIPE
            )
            _, stderr = await cmake_proc.communicate()
            if cmake_proc.returncode:
                logging.error(stderr)
                raise RuntimeError(stderr)

            make_proc = await asyncio.create_subprocess_exec(
                'make',
                f'-C{build_dir.name}',
                '--silent',
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.PIPE
            )
            _, stderr = await make_proc.communicate()
            if make_proc.returncode:
                logging.error(stderr)
                raise RuntimeError(stderr)
        finally:
            build_dir.cleanup()
            stash_dir.cleanup()

        return collector.python_package_name, cls.wrapper_class_name



import enum
import io
import logging
import os
import re


current_path = os.path.dirname(os.path.realpath(__file__))


class FixedStepSolver(str, enum.Enum):
    ODE1 = "ode1"
    ODE2 = "ode2"
    ODE3 = "ode3"
    ODE4 = "ode4"
    ODE5 = "ode5"
    ODE8 = "ode8"
    ODE14X = "ode14x"
    ODE1BE = "ode1be"


def export_model(slx_path: str, output_dir: str, step_size: float, solver: FixedStepSolver):
    import matlab.engine

    engine = None
    stdout_buffer = io.StringIO()
    stderr_buffer = io.StringIO()
    try:
        engine = matlab.engine.start_matlab(option="-nodesktop -nodisplay")
        engine.addpath(os.path.join(current_path, 'scripts'))
        if engine.is_incompatible(slx_path, nargout=1):
            raise ValueError(
                f"Model version was created with newer Matlab Release. " +
                f"It should be compatible with '{engine.get_current_release(nargout=1)}'."
            )
        engine.codegen(slx_path, output_dir, step_size, solver.value, nargout=0, stdout=stdout_buffer, stderr=stderr_buffer)
    except ValueError:
        raise
    except:
        error_stack = stderr_buffer.getvalue()
        logging.error(stdout_buffer.getvalue() + error_stack)
        cause_block = extract_cause_block(error_stack)
        if not cause_block:
            cause_block = extract_error_in_block(error_stack)
        raise ValueError(cause_block) if cause_block else ValueError("Can't build the model")
    finally:
        if engine:
            engine.quit()


def extract_cause_block(stack: str) -> str | None:
    result = re.search(r"Caused by:\n\s+.*(\n(?!\s*$).*)+", stack)
    return result.group(0) if result else None


def extract_error_in_block(stack: str) -> str | None:
    result = re.search(r"Error in '.*?': (.*?)(?=\n|$)", stack)
    return result.group(0) if result else None



from pycparser import c_ast as ast

from .Node import Node


class FunctionNodeParameter(Node):
    def __init__(self, param):
        param.type, _ = self._skip_ptr_decls(param.type)
        self.name = param.name
        self.type = " ".join(param.type.type.names)

    @staticmethod
    def is_type(param):
        return (isinstance(param, ast.Typename) or isinstance(param, ast.Decl)) and param.name is not None



import abc

from pycparser import c_ast as ast


class Node(metaclass=abc.ABCMeta):
    @staticmethod
    @abc.abstractmethod
    def is_type(node):
        return

    @staticmethod
    def _skip_ptr_decls(node, lvl=0):
        if isinstance(node, ast.PtrDecl):
            return Node._skip_ptr_decls(node.type, lvl+1)
        return node, lvl

